"""
Contains a `ProcessRequest` class to process and get a response
from a given route and requested method. Use to process client requests.
"""

import http_pyparser
from typing import List, Tuple

from .route import Route
from .status import status
from .server import Client


class RequestData:
    def __init__(
        self,
        parsed_http: http_pyparser.parser.HTTPData,
        remote_addr: tuple,
        parameters: dict
    ):
        self.real_path = parsed_http.real_path

        self.path = parsed_http.path
        self.method = parsed_http.method
        self.version = parsed_http.version
        
        self.host = parsed_http.host
        self.user_agent = parsed_http.user_agent
        self.accept = parsed_http.accept

        self.body = parsed_http.body
        self.headers = parsed_http.headers
        self.cookies = parsed_http.cookies
        self.query = parsed_http.query

        self.remote_addr = remote_addr
        self.parameters = parameters

    def __repr__(self) -> str:
        return (f'RequestData(real_path="{self.real_path}", path="{self.path}", method="{self.method}", '
                f'version="{self.version}", host="{self.host}", user_agent="{self.user_agent}", '
                f'accept="{self.accept}", body={self.body}, headers={self.headers}, cookies={self.cookies}, '
                f'query={self.query}, remote_addr={self.remote_addr}, parameters={self.parameters})')


class ProcessRequest:
    def __init__(self, routes: List[Route]) -> None:
        self._routes = routes

    def process(self, client: Client) -> Tuple[http_pyparser.Response, RequestData]:
        """Process and get or create a response to
        specified path and requested method.

        The HTTP message is obtained by `Client`
        class; the `http_pyparser` library parse this
        and return a class with HTTP data.

        If successful, this method will return an instance
        of `http_pyparser.Response` with the response generated
        to the client and an instance of `RequestData` with the
        request data.

        :param client: A `Client` instance
        :type client: Client
        :return: Return response and request object
        :rtype: Tuple[http_pyparser.Response, RequestData]
        """

        # get client request
        message = client.get_message()

        if not message:
            client.destroy()
        else:
            http_parser = http_pyparser.parser.HTTPParser()
            parsed_http = http_parser.parser(message)

            match_route: Route = None

            # checking routes
            for route in self._routes:
                parameters = route.get_parameters(parsed_http.path)

                if route.match_route(parsed_http.path) or parameters is not False:
                    match_route = route
                    break

            remote_addr = client.get_address()
            request = RequestData(parsed_http, remote_addr, parameters)

            # make route response
            if match_route:
                if route.accept_method(request.method):
                    response = route.get_route_response(request)
                else:
                    response = http_pyparser.response.Response(
                        body='Method Not Allowed',
                        status=status.method_not_allowed_405
                    )
            else:
                response = http_pyparser.response.Response(
                    body='Not Found',
                    status=status.not_found_404
                )

            return response, request



"""
The `Server` class creates and manipulates the `socket`
of the server. The `Client` class, on the other hand, is
used to store and manipulate information from the client socket.
"""

import socket
from typing import Union


class Client:
    def __init__(self, client: socket.socket, address: tuple) -> None:
        self._client = client
        self._address = address

    def send_message(self, message: str) -> None:
        """Send message to client.

        :param message: Message to send
        :type message: str
        """

        self._client.send(message.encode())

    def get_address(self) -> tuple:
        """Get client address in a tuple (host and port).

        :return: Client address.
        :rtype: tuple
        """

        return self._address

    def get_message(self) -> Union[None, str]:
        """Get the message sent by the client socket.

        :return: Client message.
        :rtype: Union[None, str]
        """

        try:
            self._client.settimeout(1)
            message = self._client.recv(1024)
        except socket.timeout:
            return None

        self._client.settimeout(None)
        return message.decode()

    def destroy(self) -> None:
        """Destroy client connection."""

        self._client.shutdown(socket.SHUT_RDWR)
        self._client.close()


class Server(object):
    def __init__(self):
        """Create a socket in TCP protocol.

        The socket is configured to reuse the address
        that is passed in the `self.start` method
        """

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def start(self, host: str, port: int, max_listen: int = 128) -> None:
        """Start socket in specified address.

        The `max_listen` argument is the maximum
        number of connections the socket supports,
        the default is 128.

        :param host: Host
        :type host: str
        :param port: Port
        :type port: int
        :param max_listen: maximum number of
        connections the socket, defaults to 128
        :type max_listen: int, optional
        """

        address = (host, port)

        self._socket.bind(address)
        self._socket.listen(max_listen)

    def wait_client(self) -> Client:
        """Wait a client connection.

        :return: client informations
        :rtype: Client
        """

        csocket, address = self._socket.accept()
        client = Client(csocket, address)

        return client



from setuptools import setup
from wsblib import __version__

with open('README.md', 'r') as reader:
    readme = reader.read()

setup(
    author='Jaedson Silva',
    author_email='imunknowuser@protonmail.com',
    name='wsblib',
    version=__version__,
    description='Base library for other web servers.',
    long_description_content_type='text/markdown',
    long_description=readme,
    packages=['wsblib'],
    install_requires=['http_pyparser==0.4.1'],
    url='https://github.com/firlast/wsblib',
    keywords=['www', 'http', 'server', 'internet', 'socket'],
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: WSGI'
    ],
    project_urls={
        'License': 'https://github.com/firlast/wsblib/blob/master/LICENSE'
    }
)


"""
The `HTTPStatus` class stores HTTP status codes from your message.
Use the instance of this class to get the status codes.
"""


class HTTPStatus:
    continue_100 = 100
    switching_protocol_101 = 101
    processing_102 = 102
    early_hints_103 = 103

    ok_200 = 200
    created_201 = 201
    accepted_202 = 202
    non_authoritative_information_203 = 203
    no_content_204 = 204
    reset_content_205 = 205
    partial_content_206 = 206
    multi_status_207 = 207
    im_used_226 = 226

    multiple_choice_300 = 300
    moved_permanently_301 = 301
    found_302 = 302
    see_other_303 = 303
    not_modified_304 = 304
    temporary_redirect_307 = 307
    permanent_redirect_308 = 308

    bad_request_400 = 400
    unauthorized_401 = 401
    payment_required_402 = 402
    forbidden_403 = 403
    not_found_404 = 404
    method_not_allowed_405 = 405
    not_acceptable_406 = 406
    proxy_authentication_required_407 = 407
    request_timeout_408 = 408
    conflict_409 = 409
    gone_410 = 410
    length_required_411 = 411
    precondition_failed_412 = 412
    payload_too_large_413 = 413
    uri_too_long_414 = 414
    unsupported_media_type_415 = 415
    requested_range_not_satisfiable_416 = 416
    expectation_failed_417 = 417
    im_a_teapot_418 = 418
    misdirected_request_421 = 421
    unprocessable_entity_422 = 422
    locked_423 = 423
    failed_dependency_424 = 424
    too_early_425 = 425
    upgrade_required_426 = 426
    precondition_required_428 = 428
    too_many_requests_429 = 429
    request_header_fields_too_large_431 = 431
    unavailable_for_legal_reasons_451 = 451

    internal_server_error_500 = 500
    not_implemented_501 = 501
    bad_gateway_502 = 502
    service_unavailable_503 = 503
    gateway_timeout_504 = 504
    http_version_not_supported_505 = 505
    variant_also_negotiates_506 = 506
    insufficient_storage_507 = 507
    loop_detected_508 = 508
    not_extended_510 = 510
    network_authentication_required_511 = 511


status = HTTPStatus()



"""
Contains a `Route` class that stores and manipulates information
of a created route, such as path, accepted methods and the callback function.
"""

from types import FunctionType
from typing import Union
from http_pyparser import response, parser

from .exceptions import InvalidRouteResponseError


class Route:
    def __init__(self, callback: FunctionType, path: str, methods: tuple = ('GET',)) -> None:
        """Creating a new `Route` instance.

        This class is used to manipulate and obtain data
        from the route requested by the user, being able to check
        methods, obtain parameters (for dynamic routes) and others.

        :param callback: Function to be called when the route is requested
        :type callback: FunctionType
        :param path: Route path
        :type path: str
        :param methods: Route allowed methods, defaults to ('GET',)
        :type methods: tuple, optional
        """

        self._path = path
        self._methods = methods
        self._callback = callback

        self._parameters = []
        self._no_parameters = []

        if '<' in path and '>' in path:
            self._register_dynamic_route()

    def _register_dynamic_route(self) -> None:
        split_path = self._path.split('/')        
        split_path.remove('')

        parameters = []
        no_parameters = []

        for index, i in enumerate(split_path):
            if i.startswith('<') and i.endswith('>'):
                i = i.replace('<', '')
                i = i.replace('>', '')
                i = i.replace(' ', '')

                parameter = i.split(':')

                if len(parameter) == 2:
                    var_type, name = parameter
                else:
                    var_type = 'any'
                    name = i

                parameters.append({'index': index, 'var_type': var_type, 'name': name})
            else:
                no_parameters.append({'index': index, 'name': i})

        # register registering dynamic route variables
        self._parameters = parameters
        self._no_parameters = no_parameters

    def _get_route_parameters(self, path_split) -> dict:
        result = {}

        for d in self._parameters:
            index = d['index']
            var_type = d['var_type']
            name = d['name']

            variable = path_split[index]
            
            if var_type == 'str':
                variable = str(variable)
            elif var_type == 'int':
                variable = int(variable)
            elif var_type == 'float':
                variable = float(variable)

            result[name] = variable

        return result

    def get_parameters(self, requested_path: str) -> Union[dict, bool]:
        """Gets the route path parameters.

        If the route is dynamic, this method will get
        the parameters and return them in a dictionary,
        if the route does not match the original route,
        `False` is returned.

        :param requested_path: Path requested by client;
        :type requested_path: str
        :return: Return the parameters or False if route does
        not match the original route.
        :rtype: Union[dict, bool]
        """

        path_split = requested_path.split('/')

        while '' in path_split:
            path_split.remove('')

        if self._path == requested_path:
            return {}
        elif len(path_split) == (len(self._parameters) + len(self._no_parameters)):
            # checking if routes that are not parameters are available
            for a in self._no_parameters:
                if path_split[a['index']] == a['name']:
                    return self._get_route_parameters(path_split)
                else:
                    break

        return False

    def match_route(self, path: str) -> bool:
        """Checks if the path specified by the "path"
        argument is the same as the path of the registered route.

        :param path: Path to check
        :type path: str
        :return: Comparison result
        :rtype: bool
        """

        return self._path == path

    def accept_method(self, method: str) -> bool:
        """Checks if the method passed by the
        argument is accepted by the route.

        :param method: HTTP method
        :type method: str
        :return: Check result
        :rtype: bool
        """

        return method in self._methods

    def get_route_response(self, request: parser.HTTPData) -> response.Response:
        """Gets the return of the route's callback
        function to use as the route's response.

        The `request` argument is only passed as
        an argument to the callback function, if it
        requests, so that it can get data from the request.

        :param request: Request data
        :type request: parser.HTTPData
        :raises InvalidRouteResponseError: If the route returns None,
        or a boolean value.
        :return: Route response in Response object;
        :rtype: response.Response
        """

        try:
            callback_response = self._callback.__call__(request)
        except TypeError:
            callback_response = self._callback.__call__()

        if not callback_response:
            raise InvalidRouteResponseError(f'Route "{self._path}" returned a invalid response')
        else:
            if isinstance(callback_response, tuple):
                # getting body and status of response
                # in use cases of: return "Hello", 200.
                body, status = callback_response
                final_response = response.Response(body, status=status)
            elif isinstance(callback_response, response.Response):
                final_response = callback_response
            else:
                final_response = response.Response(callback_response)

            return final_response
